@@ -1,4 +1,4 @@
-/* Copyright (C) 1999, 2000 Chris Vaill
+/* Copyright (C) 1999--2001 Chris Vaill
    This file is part of normalize.
 
    This program is free software; you can redistribute it and/or modify
@@ -81,6 +81,7 @@ extern int gethostname();
 
 #include "riff.h"
 
+#define USE_TEMPFILE 0
 
 #ifndef MIN
 # define MIN(a,b) ((a)<(b)?(a):(b))
@@ -122,20 +123,19 @@ struct progress_struct {
   int on_file;         /* the index of the file we're working on */
 } progress_info;
 
+void compute_levels(double *levels, char **fnames, int nfiles);
 double analyze_levels(double *levels, int nfiles, double threshold);
 double signal_max_power(int fd, char *filename);
+double signal_max_power_stream(FILE *in, struct wavfmt *fmt, char *filename);
 int apply_gain(int read_fd, int write_fd, char *filename, double gain);
+int tmpfile_same_dir(const char *path);
+int xmkstemp(char *template);
+int xrename(const char *oldpath, const char *newpath);
 void *xmalloc(size_t size);
 
 extern char version[];
 char *progname;
 
-#define VERBOSE_QUIET    0
-#define VERBOSE_PROGRESS 1
-#define VERBOSE_INFO     2
-#define VERBOSE_DEBUG    3
-int verbose = VERBOSE_PROGRESS;
-int do_apply_gain = TRUE;
 
 void
 usage()
@@ -149,7 +149,8 @@ Normalize volume of multiple WAV files\n\
   -n, --no-adjust      compute and output the volume adjustment, but\n\
                          don't apply it to any of the files\n\
   -g, --gain=ADJ       don't compute the volume adjustment, just apply\n\
-                         adjustment ADJ to the files\n\
+                         adjustment ADJ to the files.  Use the suffix \"dB\"\n\
+                         to indicate a gain in decibels.\n\
   -b, --batch          batch mode: average the levels of all files given\n\
                          on the command line, and use one adjustment, based\n\
                          on the average level, for all files\n\
@@ -168,22 +169,34 @@ Normalize volume of multiple WAV files\n\
 Report bugs to <cvaill@cs.columbia.edu>.\n"), progname);
 }
 
+enum verbose_t {
+  VERBOSE_QUIET    = 0,
+  VERBOSE_PROGRESS = 1,
+  VERBOSE_INFO     = 2,
+  VERBOSE_DEBUG    = 3,
+};
+
+/* arguments */
+int verbose = VERBOSE_PROGRESS;
+int do_print_only = FALSE;
+int do_apply_gain = TRUE;
+double target = 0.25;
+double threshold = -1.0; /* in decibels */
+int do_compute_levels = TRUE;
+int gain_in_decibels = FALSE;
+int batch_mode = FALSE;
+int mix_mode = FALSE;
+
 int
 main(int argc, char *argv[])
 {
   int fd, fd2, c, i, nfiles;
   double *levels;
-  double level, gain = 1.0;
-  char **fnames;
+  double level, rmsamp, gain = 1.0, dBdiff;
+  char **fnames, *p;
+  struct wavfmt fmt = { 1, 2, 44100, 176400, 0, 16 };
   struct stat st;
 
-  /* arguments */
-  double target = 0.25;
-  double threshold = -1.0; /* in decibels */
-  int do_compute_levels = TRUE;
-  int normalize_separately = TRUE;
-  int mix_mode = FALSE;
-
   struct option longopts[] = {
     {"help", 0, NULL, 'h'},
     {"version", 0, NULL, 'V'},
@@ -194,7 +207,7 @@ main(int argc, char *argv[])
     {"amplitude", 1, NULL, 'a'},
     {"threshold", 1, NULL, 't'},
     {"gain", 1, NULL, 'g'},
-    {"mix", 1, NULL, 'm'},
+    {"mix", 0, NULL, 'm'},
     {NULL, 0, NULL, 0}
   };
 
@@ -227,19 +240,32 @@ main(int argc, char *argv[])
       threshold = strtod(optarg, NULL);
       break;
     case 'g':
-      gain = strtod(optarg, NULL);
+      gain = strtod(optarg, &p);
+
+      /* check if "dB" is given after number */
+      while(isspace(*p))
+	p++;
+      for (i = 0; p[i]; i++)
+	p[i] = tolower(p[i]);
+      if (strncmp(p, "db", 2) == 0) {
+	dBdiff = gain;
+	gain = pow(10, dBdiff / 20);
+	gain_in_decibels = TRUE;
+      }
+
       do_compute_levels = FALSE;
-      normalize_separately = FALSE;
+      batch_mode = TRUE;
       if (gain < 0) {
 	usage();
 	exit(1);
       }
       break;
     case 'n':
+      do_print_only = TRUE;
       do_apply_gain = FALSE;
       break;
     case 'b':
-      normalize_separately = FALSE;
+      batch_mode = TRUE;
       break;
     case 'm':
       mix_mode = TRUE;
@@ -261,7 +287,7 @@ main(int argc, char *argv[])
       exit(1);
     }
   }
-  if (mix_mode && !normalize_separately) {
+  if (mix_mode && batch_mode) {
     fprintf(stderr,
 	    _("%s: error: the -m and -b options are mutually exclusive\n"),
 	    progname);
@@ -281,7 +307,13 @@ main(int argc, char *argv[])
   fnames = (char **)xmalloc((argc - optind) * sizeof(char *));
   progress_info.file_sizes = (off_t *)xmalloc((argc - optind) * sizeof(off_t));
   for (i = optind; i < argc; i++) {
-    if (stat(argv[i], &st) == -1) {
+    if (strcmp(argv[i], "-") == 0) {
+      if (do_apply_gain) {
+	fprintf(stderr, _("%s: Warning: stdin specified on command line, not adjusting files\n"), progname);
+	do_apply_gain = FALSE;
+      }
+      fnames[nfiles++] = argv[i];
+    } else if (stat(argv[i], &st) == -1) {
       fprintf(stderr, _("%s: file %s: %s\n"),
 	      progname, argv[i], strerror(errno));
     } else {
@@ -299,21 +331,83 @@ main(int argc, char *argv[])
     return 1;
   }
 
-
-  /*
-   * allocate space to store levels
-   */
+  /* allocate space to store levels */
   levels = (double *)xmalloc(nfiles * sizeof(double));
 
 
   /*
-   * Compute the power levels of the files, then figure out the
-   * necessary gain to get the power value to the target.
+   * Compute the levels
    */
   if (do_compute_levels) {
+    compute_levels(levels, fnames, nfiles);
 
-    if (verbose >= VERBOSE_PROGRESS)
-      fprintf(stderr, _("Computing levels...\n"));
+    if (batch_mode || mix_mode) {
+      level = analyze_levels(levels, nfiles, threshold);
+
+      /* level refers to signal power, but target is an amplitude, so we
+	 take the square root of the power to get the RMS amplitude */
+      rmsamp = sqrt(level);
+
+      /* For mix mode, we set the target to the average level */
+      if (mix_mode)
+	target = rmsamp;
+
+      /* For batch mode, we use one gain for all files */
+      if (batch_mode)
+	gain = target / rmsamp;
+
+      if (do_print_only)
+	printf(_("%f\taverage level\n"), rmsamp);
+      else if (verbose >= VERBOSE_INFO)
+	fprintf(stderr, _("Average level: %0.4f\n"), rmsamp);
+    }
+
+  } /* end of if (do_compute_levels) */
+
+
+  /*
+   * Check if we need to apply the gain --
+   *
+   *   Volume differences of less than 1 dB are not audible.  We
+   *   therefore allow amplitudes that are +/-0.25 dB from the target
+   *   to pass without adjustment (mainly so that normalizing files
+   *   for the second time has no effect).
+   *
+   *   Why 0.25 dB?  If we allow amplitudes that are 0.25 dB above and
+   *   below the target, the total possible range is 0.5 dB, which is
+   *   well under the 1 dB hearing threshold.
+   */
+  if (do_apply_gain && batch_mode) {
+    if (!gain_in_decibels)
+      dBdiff = 20 * log10(gain);
+
+    /* the do_compute_levels check makes sure we always apply gain if
+       it was specified with -g */
+    if (do_compute_levels && fabs(dBdiff) < 0.25) {
+      if (verbose >= VERBOSE_PROGRESS)
+	fprintf(stderr,
+		_("Files are already normalized, not adjusting...\n"));
+      do_apply_gain = FALSE;
+    } else if (verbose >= VERBOSE_PROGRESS) {
+      if (!do_compute_levels) { /* if -g */
+	if (gain_in_decibels)
+	  printf(_("Applying adjustment of %fdB...\n"), dBdiff);
+	else
+	  printf(_("Applying adjustment of %f...\n"), gain);
+      } else {
+	printf(_("Applying adjustment of %0.2fdB...\n"), dBdiff);
+      }
+    }
+  }
+
+
+  /*
+   * Apply the gain
+   */
+  if (do_apply_gain) {
+    struct stat stbuf;
+    char *tmpfile, *p;
+    int tmpnum;
 
     progress_info.batch_start = time(NULL);
     progress_info.finished_size = 0;
@@ -327,215 +421,194 @@ main(int argc, char *argv[])
 	continue;
       }
 
+      if (!batch_mode) {
+	gain = target / sqrt(levels[i]);
+	dBdiff = 20 * log10(gain);
+
+	/* don't bother applying very small adjustments -- see above */
+	if (fabs(dBdiff) < 0.25) {
+	  if (verbose >= VERBOSE_PROGRESS)
+	    fprintf(stderr, _("%s already normalized, not adjusting...\n"),
+		    fnames[i]);
+	  continue;
+	}
+
+	if (verbose >= VERBOSE_PROGRESS)
+	  fprintf(stderr, _("Applying adjustment of %0.2fdB to %s...\n"),
+		  dBdiff, fnames[i]);
+      }
+
+#if USE_TEMPFILE
+      /* Create temporary file name, and open it for writing.  We want
+       * it to be in the same directory (and therefore, in the same
+       * filesystem, for a fast rename). */
+      tmpfile = (char *)xmalloc(strlen(fnames[i]) + 16);
+      strcpy(tmpfile, fnames[i]);
+      if ((p = strrchr(tmpfile, '/')) == NULL)
+	p = tmpfile;
+      else
+	p++;
+      *p = '\0';
+      strcpy(p, "_normXXXXXX");
+      fd2 = xmkstemp(tmpfile);
+      if (fd2 == -1) {
+	fprintf(stderr, _("%s: error opening temp file: %s\n"), progname,
+		strerror(errno));
+	close(fd);
+	free(tmpfile);
+	continue;
+      }
+
+      /* preserve original permissions */
+      fstat(fd, &stbuf);
+      fchmod(fd2, stbuf.st_mode);
+#else
+      fd2 = open(fnames[i], O_WRONLY);
+      if (fd2 == -1) {
+	fprintf(stderr, _("%s: error opening %s: %s\n"), progname, fnames[i],
+		strerror(errno));
+	close(fd);
+	continue;
+      }
+#endif
+
+      progress_info.file_start = time(NULL);
+      progress_info.on_file = i;
+
+      if (apply_gain(fd, fd2, fnames[i], gain) == -1) {
+	fprintf(stderr, _("%s: error applying adjustment to %s: %s\n"),
+		progname, fnames[i], strerror(errno));
+      }
+
+      progress_info.finished_size += progress_info.file_sizes[i];
+
+      close(fd);
+      close(fd2);
+
+#if USE_TEMPFILE
+      /* move the temporary file back to the original file */
+      if (xrename(tmpfile, fnames[i]) == -1) {
+	fprintf(stderr, _("%s: error moving %s to %s: %s\n"), progname,
+		tmpfile, fnames[i], strerror(errno));
+	exit(1);
+      }
+      free(tmpfile);
+#endif
+
+      if (verbose >= VERBOSE_PROGRESS && !batch_mode)
+	fprintf(stderr, "\n");
+    }
+
+    /* we're done with the second progress meter, so go to next line */
+    if (verbose >= VERBOSE_PROGRESS && batch_mode)
+      fputc('\n', stderr);
+
+  } else if (batch_mode && do_print_only) {
+
+    /* if we're not applying the gain, just print it out, and we're done */
+    dBdiff = 20 * log10(gain);
+    printf(_("%fdB\tvolume adjustment\n"), dBdiff);
+
+  }
+
+  free(levels);
+  free(progress_info.file_sizes);
+  free(fnames);
+
+  return 0;
+}
+
+void
+compute_levels(double *levels, char **fnames, int nfiles)
+{
+  double level, gain = 1.0;
+  int i, fd;
+  struct wavfmt fmt = { 1, 2, 44100, 176400, 0, 16 };
+
+  /*
+   * Compute the power levels of the files, then figure out the
+   * necessary gain to get the power value to the target.
+   */
+  if (verbose >= VERBOSE_PROGRESS)
+    fprintf(stderr, _("Computing levels...\n"));
+
+  progress_info.batch_start = time(NULL);
+  progress_info.finished_size = 0;
+
+  for (i = 0; i < nfiles; i++) {
+
+    levels[i] = 0;
+
+    if (strcmp(fnames[i], "-") == 0) {
+      progress_info.file_start = time(NULL);
+      progress_info.on_file = i;
+      errno = 0;
+
+      level = signal_max_power_stream(stdin, &fmt, NULL);
+      fnames[i] = "STDIN";
+
+    } else {
+
+      fd = open(fnames[i], O_RDONLY);
+      if (fd == -1) {
+	fprintf(stderr, _("%s: error opening %s: %s\n"), progname, fnames[i],
+		strerror(errno));
+	continue;
+      }
+
       progress_info.file_start = time(NULL);
       progress_info.on_file = i;
       errno = 0;
+
       level = signal_max_power(fd, fnames[i]);
-      if (level < -EPSILON) {
-	fprintf(stderr, _("%s: error reading %s: %s\n"), progname, fnames[i],
-		strerror(errno));
-	goto error1_1;
-      }
-      if (level < EPSILON) {
-	if (verbose >= VERBOSE_PROGRESS) {
-	  fprintf(stderr,
-		  "\r                                     "
-		  "                                     \r");
-	  fprintf(stderr,
-		  _("File %s has zero power, ignoring...\n"), fnames[i]);
-	}
-	goto error1_1;
-      }
+    }
 
-      progress_info.finished_size += progress_info.file_sizes[i];
-
-      levels[i] = level;
-      if (!do_apply_gain) {
-
-	/* clear the progress meter first */
-	if (verbose >= VERBOSE_PROGRESS)
-	  fprintf(stderr,
-		  "\r                                     "
-		  "                                     \r");
-
-	printf("%f\t%s\n", sqrt(levels[i]), fnames[i]);
-
-      } else if (verbose >= VERBOSE_INFO) {
+    if (level < -EPSILON) {
+      fprintf(stderr, _("%s: error reading %s: %s\n"), progname, fnames[i],
+	      strerror(errno));
+      goto error_close_fd;
+    }
+    if (level < EPSILON) {
+      if (verbose >= VERBOSE_PROGRESS) {
 	fprintf(stderr,
 		"\r                                     "
 		"                                     \r");
-	fprintf(stderr, _("Level for %s: %0.4f\n"),
-		fnames[i], sqrt(levels[i]));
+	fprintf(stderr,
+		_("File %s has zero power, ignoring...\n"), fnames[i]);
       }
-
-    error1_1:
-      close(fd);
+      goto error_close_fd;
     }
 
-    /* we're done with the level calculation progress meter, so go to
-       next line */
-    if (verbose == VERBOSE_PROGRESS && do_apply_gain)
-      fputc('\n', stderr);
+    levels[i] = level;
+    if (do_print_only) {
 
-    if (!normalize_separately || mix_mode) {
-      level = analyze_levels(levels, nfiles, threshold);
+      /* clear the progress meter first */
+      if (verbose >= VERBOSE_PROGRESS)
+	fprintf(stderr,
+		"\r                                     "
+		"                                     \r");
 
-      if (!do_apply_gain)
-	printf(_("%f\taverage level\n"), sqrt(level));
-      else if (verbose >= VERBOSE_INFO)
-	printf(_("Average level: %f\n"), sqrt(level));
+      printf("%f\t%s\n", sqrt(levels[i]), fnames[i]);
 
-      /* level refers to signal power, but target is an amplitude, so we
-	 take the square root of the power to get the RMS amplitude */
-      gain = target / sqrt(level);
-    } /* end of if (!normalize_separately) */
-
-    /*
-     * For mix mode, we set the target to the average level
-     */
-    if (mix_mode)
-      target = sqrt(level);
-
-  } /* end of if (do_compute_levels) */
-
-
-  /*
-   * Apply the gain
-   */
-  if (do_apply_gain) {
-    struct stat stbuf;
-    char *tmpfile, *p;
-    int tmpnum;
-
-    if (verbose >= VERBOSE_PROGRESS && !normalize_separately)
-      printf(_("Applying adjustment of %0.3f...\n"), gain);
-
-    progress_info.batch_start = time(NULL);
-    progress_info.finished_size = 0;
-
-    for (i = 0; i < nfiles; i++) {
-
-      fd = open(fnames[i], O_RDONLY);
-      if (fd == -1) {
-	fprintf(stderr, _("%s: error opening %s: %s\n"), progname, fnames[i],
-		strerror(errno));
-	continue;
-      }
-#if 1
-      fstat(fd, &stbuf);
-      /* Create temporary file name, and open it for writing.  We
-       * don't use tmpnam() because we want it to be in the same
-       * directory (and therefore, in the same filesystem, for a fast
-       * rename). */
-      tmpfile = (char *)xmalloc(strlen(fnames[i]) + 32);
-      strcpy(tmpfile, fnames[i]);
-      if ((p = strrchr(tmpfile, '/')) == NULL) {
-	tmpfile[0] = '\0';
-	p = tmpfile;
-      } else {
-	p++;
-	*p = '\0';
-      }
-      strcat(p, "_norm"); p += 5;
-      gethostname(p, 20);
-      p = tmpfile + strlen(tmpfile);
-      tmpnum = getpid();
-      do {
-	sprintf(p, "%d", tmpnum);
-	fd2 = open(tmpfile, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC,
-		   stbuf.st_mode);
-	tmpnum++;
-      } while (fd2 == -1 && errno == EEXIST);
-      if (fd2 == -1) {
-	fprintf(stderr, _("%s: error opening temp file: %s\n"), progname,
-		strerror(errno));
-	close(fd);
-	free(tmpfile);
-	continue;
-      }
-#else
-      fd2 = open(fnames[i], O_WRONLY);
-      if (fd2 == -1) {
-	fprintf(stderr, _("%s: error opening %s: %s\n"), progname, fnames[i],
-		strerror(errno));
-	close(fd);
-	continue;
-      }
-#endif
-
-      if (normalize_separately)
-	gain = target / sqrt(levels[i]);
-
-      if (verbose >= VERBOSE_PROGRESS && normalize_separately)
-	fprintf(stderr, _("Applying adjustment of %0.3f to %s...\n"),
-		gain, fnames[i]);
-
-      progress_info.file_start = time(NULL);
-      progress_info.on_file = i;
-      if (apply_gain(fd, fd2, fnames[i], gain) == -1) {
-	fprintf(stderr, _("%s: error applying adjustment to %s: %s\n"),
-		progname, fnames[i], strerror(errno));
-      }
-
-      progress_info.finished_size += progress_info.file_sizes[i];
-
-      close(fd);
-      close(fd2);
-
-#if 1
-      /*
-       * Move the temporary file back to the original file,
-       * copying if they are on different filesystems.
-       */
-      if (rename(tmpfile, fnames[i]) == -1) {
-	if (errno == EXDEV) {
-	  /* temp file is on a different filesystem, so we have to copy it */
-	  FILE *in, *out;
-	  in = fopen(tmpfile, "rb");
-	  out = fopen(fnames[i], "wb");
-	  if (in == NULL || out == NULL) {
-	    fprintf(stderr, _("%s: error moving %s to %s: %s\n"), progname,
-		    tmpfile, fnames[i], strerror(errno));
-	    exit(1);
-	  }
-	  while ((c = getc(in)) != EOF)
-	    putc(c, out);
-	  if (fclose(out) == EOF || fclose(in) == EOF) {
-	    fprintf(stderr, _("%s: error moving %s to %s: %s\n"), progname,
-		    tmpfile, fnames[i], strerror(errno));
-	    exit(1);
-	  }
-	  if (unlink(tmpfile) == -1)
-	    fprintf(stderr, _("%s: error removing %s: %s\n"), progname,
-		    tmpfile, strerror(errno));
-	} else {
-	  fprintf(stderr, _("%s: error moving %s to %s: %s\n"), progname,
-		  tmpfile, fnames[i], strerror(errno));
-	}
-      }
-      free(tmpfile);
-#endif
-
-      if (verbose >= VERBOSE_PROGRESS && normalize_separately)
-	fprintf(stderr, "\n");
+    } else if (verbose >= VERBOSE_INFO) {
+      fprintf(stderr,
+	      "\r                                     "
+	      "                                     \r");
+      fprintf(stderr, _("Level for %s: %0.4f\n"),
+	      fnames[i], sqrt(levels[i]));
     }
 
-    /* we're done with the second progress meter, so go to next line */
-    if (verbose >= VERBOSE_PROGRESS && !normalize_separately)
-      fputc('\n', stderr);
-
-  } else if (!normalize_separately) {
-
-    /* if we're not applying the gain, just print it out, and we're done */
-    printf(_("%f\tvolume adjustment\n"), gain);
+  error_close_fd:
+    progress_info.finished_size += progress_info.file_sizes[i];
 
+    if (strcmp(fnames[i], "STDIN") != 0)
+      close(fd);
   }
 
-  free(levels);
-  free(progress_info.file_sizes);
-  free(fnames);
-
-  return 0;
+  /* we're done with the level calculation progress meter, so go to
+     next line */
+  if (verbose == VERBOSE_PROGRESS && !do_print_only)
+    fputc('\n', stderr);
 }
 
 /*
@@ -586,7 +659,7 @@ analyze_levels(double *levels, int nlevels, double threshold)
     else
       std_dev = sqrt(variance);
     if (verbose >= VERBOSE_INFO)
-      printf(_("Standard deviation is %0.2f dB\n"), std_dev);
+      fprintf(stderr, _("Standard deviation is %0.2f dB\n"), std_dev);
 
     threshold = 2 * std_dev;
   }
@@ -610,7 +683,8 @@ analyze_levels(double *levels, int nlevels, double threshold)
       /* mark as bad any level that is > threshold different than the mean */
       if (level_difference > threshold) {
 	if (verbose >= VERBOSE_INFO)
-	  printf(_("Throwing out level of %0.3f (different by %0.2fdB)\n"),
+	  fprintf(stderr,
+		  _("Throwing out level of %0.4f (different by %0.2fdB)\n"),
 		 sqrt(levels[i]), level_difference);
 	badlevels[i] = TRUE;
       }
@@ -649,6 +723,15 @@ progress_callback(char *prefix, float fraction_completed)
   float batch_fraction;
   unsigned int batch_eta_hr, batch_eta_min, batch_eta_sec;
 
+  if (fraction_completed <= 0.0) {
+    batch_fraction = (progress_info.finished_size
+		      / (float)progress_info.batch_size);
+    fprintf(stderr,
+      _(" %s   0%% done, ETA --:--:-- (batch %3.0f%% done, ETA --:--:--)\r"),
+	    prefix, batch_fraction * 100);
+    return;
+  }
+
   now = time(NULL);
   if (fraction_completed > 1.0)
     fraction_completed = 1.0;
@@ -656,16 +739,15 @@ progress_callback(char *prefix, float fraction_completed)
   /* figure out the ETA for this file */
   file_eta_hr = file_eta_sec = file_eta_min = 0;
   time_spent = now - progress_info.file_start;
-  file_eta_sec = floor((float)time_spent / fraction_completed
-		       - (float)time_spent + 0.5);
-  while (file_eta_sec >= 60) {
-    file_eta_min++;
-    file_eta_sec -= 60;
-  }
-  while (file_eta_min >= 60) {
-    file_eta_hr++;
-    file_eta_min -= 60;
-  }
+  if (fraction_completed == 0.0)
+    file_eta_sec = 0;
+  else
+    file_eta_sec = floor((float)time_spent / fraction_completed
+			 - (float)time_spent + 0.5);
+  file_eta_min = file_eta_sec / 60;
+  file_eta_sec = file_eta_sec % 60;
+  file_eta_hr = file_eta_min / 60;
+  file_eta_min = file_eta_min % 60;
   if (file_eta_hr > 99)
     file_eta_hr = 99;
 
@@ -677,14 +759,11 @@ progress_callback(char *prefix, float fraction_completed)
   time_spent = now - progress_info.batch_start;
   batch_eta_sec = floor((float)time_spent / batch_fraction
 			- (float)time_spent + 0.5);
-  while (batch_eta_sec >= 60) {
-    batch_eta_min++;
-    batch_eta_sec -= 60;
-  }
-  while (batch_eta_min >= 60) {
-    batch_eta_hr++;
-    batch_eta_min -= 60;
-  }
+
+  batch_eta_min = batch_eta_sec / 60;
+  batch_eta_sec = batch_eta_sec % 60;
+  batch_eta_hr = batch_eta_min / 60;
+  batch_eta_min = batch_eta_min % 60;
   if (batch_eta_hr > 99)
     batch_eta_hr = 99;
 
@@ -880,6 +959,7 @@ get_smoothed_data(datasmooth_t *s)
   return smoothed;
 }
 
+
 /*
  * Get the maximum power level of the wav file
  */
@@ -909,6 +989,7 @@ signal_max_power(int fd, char *filename)
   unsigned char *data_buf = NULL;
   int filled_sz;
 
+
   riff = riff_new(fd, RIFF_RDONLY);
   if (riff == NULL) {
     fprintf(stderr, _("%s: error making riff object\n"), progname);
@@ -955,7 +1036,7 @@ signal_max_power(int fd, char *filename)
 
   in = fdopen(fd, "r");
   if (in == NULL) {
-    fprintf(stderr, _("%s: unable to fdopen: %s\n"),
+    fprintf(stderr, _("%s: failed fdopen: %s\n"),
 	    progname, strerror(errno));
     goto error7;
   }
@@ -1085,6 +1166,166 @@ signal_max_power(int fd, char *filename)
   return -1.0;
 }
 
+
+/*
+ * Get the maximum power level of the data read from a stream
+ */
+double
+signal_max_power_stream(FILE *in, struct wavfmt *fmt, char *filename)
+{
+  int bytes_per_sample;
+  int last_window;
+  unsigned int windowsz;
+  unsigned int win_start, old_start, win_end, old_end;
+
+  int i, c;
+  long sample, samplemax;
+  double *sums;
+  double pow, maxpow;
+  datasmooth_t *powsmooth;
+
+  float progress, last_progress = 0.0;
+  char prefix_buf[18];
+
+  unsigned char *data_buf = NULL;
+  int filled_sz;
+
+  if (filename == NULL || strcmp(filename, "-") == 0)
+    filename = "STDIN";
+
+  windowsz = (unsigned int)(fmt->samples_per_sec / 100);
+
+  bytes_per_sample = (fmt->bits_per_sample - 1) / 8 + 1;
+  samplemax = 1 << (bytes_per_sample * 8 - 1);
+
+  sums = (double *)xmalloc(fmt->channels * sizeof(double));
+  for (c = 0; c < fmt->channels; c++)
+    sums[c] = 0;
+
+  data_buf = (unsigned char *)xmalloc(windowsz
+				      * fmt->channels * bytes_per_sample);
+
+  /* set up smoothing window buffer */
+  powsmooth = (datasmooth_t *)xmalloc(fmt->channels * sizeof(datasmooth_t));
+  for (c = 0; c < fmt->channels; c++) {
+    powsmooth[c].buflen = 100; /* use a 100-element (1 second) window */
+    powsmooth[c].buf = (double *)xmalloc(powsmooth[c].buflen * sizeof(double));
+    powsmooth[c].start = powsmooth[c].n = 0;
+  }
+
+  /* initialize progress meter */
+  if (verbose >= VERBOSE_PROGRESS) {
+    if (strrchr(filename, '/') != NULL) {
+      filename = strrchr(filename, '/');
+      filename++;
+    }
+    strncpy(prefix_buf, filename, 17);
+    prefix_buf[17] = 0;
+    progress_callback(prefix_buf, 0.0);
+    last_progress = 0.0;
+  }
+
+
+  /*
+   * win_start, win_end, old_start, windowsz, interval, and i are in
+   * units of samples.  c is in units of channels.
+   *
+   * The actual window extends from win_start to win_end - 1, inclusive.
+   */
+  old_start = win_start = 0;
+  win_end = 0;
+  last_window = FALSE;
+  maxpow = 0.0;
+
+  do {
+
+    /* set up the window end */
+    old_end = win_end;
+    win_end = win_start + windowsz;
+
+    /* read a windowsz sized chunk */
+    filled_sz = fread(data_buf, bytes_per_sample,
+		      windowsz * fmt->channels, in);
+
+    /* if we couldn't read a complete chunk, then this is the last chunk */
+    if (filled_sz < windowsz * fmt->channels) {
+      win_end = win_start + (filled_sz / fmt->channels);
+      last_window = TRUE;
+    }
+
+    for (c = 0; c < fmt->channels; c++) {
+      sums[c] = 0;
+      /*for (i = 0; i * bytes_per_sample * fmt->channels < filled_sz; i++) {*/
+      for (i = 0; i < (win_end - win_start); i++) {
+	sample = get_sample(data_buf + (i * fmt->channels * bytes_per_sample)
+			    + (c * bytes_per_sample), bytes_per_sample);
+	sums[c] += sample * sample;
+      }
+    }
+
+    /* compute power for each channel */
+    for (c = 0; c < fmt->channels; c++) {
+      int end;
+      pow = sums[c] / (double)(win_end - win_start);
+
+      end = (powsmooth[c].start + powsmooth[c].n) % powsmooth[c].buflen;
+      powsmooth[c].buf[end] = pow;
+      if (powsmooth[c].n == powsmooth[c].buflen) {
+	powsmooth[c].start = (powsmooth[c].start + 1) % powsmooth[c].buflen;
+	pow = get_smoothed_data(&powsmooth[c]);
+	if (pow > maxpow)
+	  maxpow = pow;
+      } else {
+	powsmooth[c].n++;
+      }
+    }
+
+    /* slide the window ahead */
+    old_start = win_start;
+    win_start += windowsz;
+
+  } while (!last_window);
+
+  if (maxpow < EPSILON) {
+    /*
+     * Either this whole file has zero power, or was too short to ever
+     * fill the smoothing buffer.  In the latter case, we need to just
+     * get maxpow from whatever data we did collect.
+     */
+    for (c = 0; c < fmt->channels; c++) {
+      pow = get_smoothed_data(&powsmooth[c]);
+      if (pow > maxpow)
+	maxpow = pow;
+    }
+  }
+
+  for (c = 0; c < fmt->channels; c++)
+    free(powsmooth[c].buf);
+  free(powsmooth);
+  free(data_buf);
+  free(sums);
+
+  /* scale the pow value to be in the range 0.0 -- 1.0 */
+  maxpow = maxpow / (samplemax * samplemax);
+
+  return maxpow;
+
+  /* error handling stuff */
+ error7:
+  for (c = 0; c < fmt->channels; c++)
+    free(powsmooth[c].buf);
+  /*error6:*/
+  free(powsmooth);
+  /*error5:*/
+  free(data_buf);
+  /*error4:*/
+  free(sums);
+  /*error3:*/
+ error1:
+  return -1.0;
+}
+
+
 int
 apply_gain(int read_fd, int write_fd, char *filename, double gain)
 {
@@ -1255,7 +1496,8 @@ apply_gain(int read_fd, int write_fd, char *filename, double gain)
     }
   } else if (verbose >= VERBOSE_PROGRESS) {
     if (clip_loss > CLIPPING_WARN_THRESH)
-      fprintf(stderr, _("%s: Warning: lost %.2f%% of data due to clipping\n"),
+      fprintf(stderr,
+        _("%s: Warning: lost %.2f%% of data due to clipping               \n"),
 	      progname, clip_loss * 100);
   }
 
@@ -1276,6 +1518,82 @@ apply_gain(int read_fd, int write_fd, char *filename, double gain)
   return -1;
 }
 
+#if USE_TEMPFILE
+/*
+ * This works like the BSD mkstemp, except that we don't unlink the
+ * file, since we end up renaming it to something else.
+ */
+int
+xmkstemp(char *template)
+{
+  static char sfx[7] = "AAAAAA";
+  char *p;
+  int fd, i, done;
+
+  p = template + strlen(template) - 6;
+  if (strcmp(p, "XXXXXX") != 0) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  do {
+    strcpy(p, sfx);
+
+    /* increment the suffix */
+    done = 0; i = 5;
+    while (!done && i >= 0) {
+      sfx[i]++;
+      if (sfx[i] > 'Z') {
+	sfx[i] = 'A';
+	i--;
+      } else {
+	done = 1;
+      }
+    }
+    if (!done) {
+      errno = EEXIST;
+      return -1;
+    }
+
+    /* attempt to open the file */
+    fd = open(template, O_RDWR | O_CREAT | O_EXCL, 0600);
+
+  } while (fd == -1 && errno == EEXIST);
+
+  return fd;
+}
+
+
+/*
+ * Move the file "oldpath" to "newpath", or copy and delete if they
+ * are on different filesystems.
+ */
+int
+xrename(const char *oldpath, const char *newpath)
+{
+  if (rename(oldpath, newpath) == -1) {
+    if (errno == EXDEV) {
+      /* files are on different filesystems, so we have to copy */
+      FILE *in, *out;
+      char buf[4096];
+      size_t sz;
+      in = fopen(oldpath, "rb");
+      out = fopen(newpath, "wb");
+      if (in == NULL || out == NULL)
+	return -1;
+      while (sz = fread(buf, 1, 4096, in))
+	fwrite(buf, 1, sz, out);
+      if (fclose(out) == EOF || fclose(in) == EOF)
+	return -1;
+      if (unlink(oldpath) == -1)
+	return -1;
+    } else {
+      return -1;
+    }
+  }
+}
+#endif
+
 void *
 xmalloc(size_t size)
 {
